from rdflib import RDF, Graph, URIRef

from .models import (
    PropertyDomainViolation,
    PropertyRangeViolation,
    UndefinedClassViolation,
    UndefinedPropertyViolation,
    Violation,
)
from .retrievers import (
    get_classes_from_definitions,
    get_classes_from_instances,
    get_data_properties,
    get_object_properties,
    get_object_properties_with_domains,
    get_object_properties_with_ranges,
    get_superclasses,
)
from .utils import get_short_name, is_annotation_property


def validate(data_graph: Graph, ont_graph: Graph) -> tuple[bool, set[Violation], str]:
    """
    Validate a data graph against an ontology graph.

    It checks for coherence between the two graphs and returns validation results.

    Args:
        data_graph: An rdflib Graph object representing the data graph to be validated.
        ont_graph: An rdflib Graph object representing the ontology graph.

    Returns:
        - bool: True if there are no violations, False otherwise
        - set[Violation]: Set of violation objects found during validation
        - str: Human-readable validation report
    """
    violations: set[Violation] = {
        *_validate_undefined_class(data_graph, ont_graph),
        *_validate_undefined_property(data_graph, ont_graph),
        *_validate_object_property_domain(data_graph, ont_graph),
        *_validate_object_property_range(data_graph, ont_graph),
    }

    conforms = len(violations) == 0

    if conforms:
        report = "Validation Report\nConforms: True\nResults (0):"
    else:
        violations_list = "\n".join(f"{v.description}" for v in violations)
        report = f"Validation Report\nConforms: False\nResults ({len(violations)}):\n{violations_list}"

    return conforms, violations, report


def _validate_undefined_class(data_graph: Graph, ont_graph: Graph) -> set[Violation]:
    ontology_classes = get_classes_from_definitions(ont_graph)
    graph_classes = get_classes_from_instances(data_graph)

    undefined_classes = graph_classes - ontology_classes
    violations: set[Violation] = set()

    if undefined_classes:
        for s, _, o in data_graph:
            if o in undefined_classes:
                instance_id = get_short_name(s, data_graph)
                undefined_class = get_short_name(o, data_graph)
                violations.add(UndefinedClassViolation(instance_id=instance_id, undefined_class=undefined_class))

    return violations


def _validate_undefined_property(data_graph: Graph, ont_graph: Graph) -> set[Violation]:
    """
    Spot properties in the test graph that are not defined on the classes in the reference ontology.
    Annotation properties are excluded from validation.
    """
    defined_object_properties = get_object_properties(ont_graph)
    defined_data_properties = get_data_properties(ont_graph)
    violations: set[Violation] = set()

    for s, p, _ in data_graph:
        if p == RDF.type or is_annotation_property(p):
            continue

        if isinstance(p, URIRef) and p not in defined_object_properties and p not in defined_data_properties:
            violations.add(
                UndefinedPropertyViolation(
                    instance_id=get_short_name(s, data_graph),
                    undefined_property=get_short_name(p, data_graph),
                )
            )

    return violations


def _validate_object_property_domain(data_graph: Graph, ont_graph: Graph) -> set[Violation]:
    relations_mapped_to_allowed_domains = get_object_properties_with_domains(ont_graph)
    violations: set[Violation] = set()

    for s, p, o in data_graph:
        s_name = get_short_name(s, data_graph)
        p_name = get_short_name(p, data_graph)
        if isinstance(o, URIRef) and isinstance(s, URIRef) and isinstance(p, URIRef) and p != URIRef(RDF.type):
            s_classes = set(data_graph.objects(subject=s, predicate=RDF.type))
            for cls in list(s_classes):
                s_classes.update(get_superclasses(cls, ont_graph))

            o_classes = set(data_graph.objects(subject=o, predicate=RDF.type))
            for cls in list(o_classes):
                o_classes.update(get_superclasses(cls, ont_graph))

            if p in relations_mapped_to_allowed_domains:
                allowed_domain_classes = relations_mapped_to_allowed_domains[p]
                if not allowed_domain_classes.intersection(s_classes):
                    violations.add(
                        PropertyDomainViolation(
                            instance_id=s_name,
                            property_name=p_name,
                            invalid_type=", ".join([get_short_name(cls, data_graph) for cls in s_classes]),
                            expected_type=", ".join([
                                get_short_name(cls, data_graph) for cls in allowed_domain_classes
                            ]),
                        )
                    )

    return violations


def _validate_object_property_range(data_graph: Graph, ont_graph: Graph) -> set[Violation]:
    relations_mapped_to_allowed_ranges = get_object_properties_with_ranges(ont_graph)
    violations: set[Violation] = set()

    for s, p, o in data_graph:
        p_name = get_short_name(p, data_graph)
        o_name = get_short_name(o, data_graph)
        if isinstance(o, URIRef) and isinstance(s, URIRef) and isinstance(p, URIRef) and p != URIRef(RDF.type):
            s_classes = set(data_graph.objects(subject=s, predicate=RDF.type))
            for cls in list(s_classes):
                s_classes.update(get_superclasses(cls, ont_graph))

            o_classes = set(data_graph.objects(subject=o, predicate=RDF.type))
            for cls in list(o_classes):
                o_classes.update(get_superclasses(cls, ont_graph))

            if p in relations_mapped_to_allowed_ranges:
                allowed_range_classes = relations_mapped_to_allowed_ranges[p]
                if not allowed_range_classes.intersection(o_classes):
                    violations.add(
                        PropertyRangeViolation(
                            instance_id=o_name,
                            property_name=p_name,
                            invalid_type=", ".join([get_short_name(cls, data_graph) for cls in o_classes]),
                            expected_type=", ".join([get_short_name(cls, data_graph) for cls in allowed_range_classes]),
                        )
                    )

    return violations



from typing import Union

from rdflib import Graph, URIRef
from rdflib.namespace import OWL, RDF, RDFS, XSD
from rdflib.term import BNode, Node


def get_classes_from_definitions(ontology: Graph) -> set[URIRef]:
    class_nodes: set[Node] = set()

    # Explicit class definitions
    class_nodes.update(ontology.subjects(RDF.type, RDFS.Class))
    class_nodes.update(ontology.subjects(RDF.type, OWL.Class))

    # OWL class axioms
    class_axioms = [OWL.equivalentClass, OWL.disjointWith, OWL.complementOf, OWL.unionOf, OWL.intersectionOf, OWL.oneOf]
    for axiom in class_axioms:
        class_nodes.update(ontology.subjects(axiom, None))
        class_nodes.update(ontology.objects(None, axiom))

    # Remove literal values and non-URIRefs
    class_uri_refs: set[URIRef] = {cls for cls in class_nodes if isinstance(cls, URIRef)}

    return class_uri_refs


def get_classes_from_instances(graph: Graph) -> set[URIRef]:
    class_nodes: set[Node] = set()

    # Classes inferred from instances
    class_nodes.update(graph.objects(None, RDF.type))

    # Remove literal values and non-URIRefs
    class_uri_refs: set[URIRef] = {cls for cls in class_nodes if isinstance(cls, URIRef)}

    return class_uri_refs


def get_object_properties(graph: Graph) -> set[URIRef]:
    # Get properties explicitly defined as ObjectProperties
    object_properties = set(graph.subjects(predicate=RDF.type, object=OWL.ObjectProperty))

    # Get regular properties that have a class as their range (not a literal)
    regular_properties = set(graph.subjects(predicate=RDF.type, object=RDF.Property))
    for prop in regular_properties:
        ranges = set(graph.objects(prop, RDFS.range))
        # Only include properties whose range is a class (not a literal type)
        if any(
            range_val
            for range_val in ranges
            if isinstance(range_val, URIRef) and not str(range_val).startswith(str(XSD))
        ):
            object_properties.add(prop)

    return {prop for prop in object_properties if isinstance(prop, URIRef)}


def get_object_properties_with_domains(ontology: Graph) -> dict[URIRef, set[URIRef]]:
    object_properties_with_domains: dict[URIRef, set[URIRef]] = {}
    object_properties = get_object_properties(ontology)
    for op in object_properties:
        domains = set(ontology.objects(subject=op, predicate=RDFS.domain))
        for d in list(domains):
            if isinstance(d, BNode):
                sub_graph = ontology.cbd(d)
                linked_domains = sub_graph.objects(predicate=RDF.first)
                domains.remove(d)
                domains.update(linked_domains)
            else:
                domains.update(get_superclasses(d, ontology))
        object_properties_with_domains[op] = {d for d in domains if isinstance(d, URIRef)}
    return object_properties_with_domains


def get_object_properties_with_ranges(ontology: Graph) -> dict[URIRef, set[URIRef]]:
    object_properties_with_ranges: dict[URIRef, set[URIRef]] = {}
    object_properties = get_object_properties(ontology)
    for op in object_properties:
        ranges = set(ontology.objects(subject=op, predicate=RDFS.range))
        for d in list(ranges):
            if isinstance(d, BNode):
                sub_graph = ontology.cbd(d)
                linked_domains = sub_graph.objects(predicate=RDF.first)
                ranges.remove(d)
                ranges.update(linked_domains)
            else:
                ranges.update(get_superclasses(d, ontology))
        object_properties_with_ranges[op] = {r for r in ranges if isinstance(r, URIRef)}
    return object_properties_with_ranges


def get_data_properties(graph: Graph) -> set[URIRef]:
    # Get properties explicitly defined as DataProperties
    data_properties = set(graph.subjects(predicate=RDF.type, object=OWL.DatatypeProperty))

    # Get regular properties that have a literal/XSD type as their range
    regular_properties = set(graph.subjects(predicate=RDF.type, object=RDF.Property))
    for prop in regular_properties:
        ranges = set(graph.objects(prop, RDFS.range))
        # Only include properties whose range is a literal type
        if any(isinstance(range_val, URIRef) and str(range_val).startswith(str(XSD)) for range_val in ranges):
            data_properties.add(prop)

    return {prop for prop in data_properties if isinstance(prop, URIRef)}


def get_data_properties_with_domains(graph: Graph) -> dict[URIRef, set[URIRef]]:
    data_properties_with_domains: dict[URIRef, set[URIRef]] = {}
    data_properties = get_data_properties(graph)
    for dp in data_properties:
        domains = set(graph.objects(subject=dp, predicate=RDFS.domain))
        for d in list(domains):
            if isinstance(d, BNode):
                sub_graph = graph.cbd(d)
                linked_domains = sub_graph.objects(predicate=RDF.first)
                domains.remove(d)
                domains.update(linked_domains)
            else:
                domains.update(get_superclasses(d, graph))
        data_properties_with_domains[dp] = {d for d in domains if isinstance(d, URIRef)}
    return data_properties_with_domains


def get_superclasses(cls: Union[URIRef, Node], ontology: Graph) -> set[Union[URIRef, Node]]:
    superclasses = set()
    to_visit = [cls]
    while to_visit:
        current = to_visit.pop()
        for superclass in ontology.objects(subject=current, predicate=RDFS.subClassOf):
            if superclass not in superclasses:
                superclasses.add(superclass)
                to_visit.append(superclass)
    return superclasses



from typing import Union

from rdflib import Graph, URIRef
from rdflib.namespace import DC, DCTERMS, RDFS, SKOS
from rdflib.term import Node

ANNOTATION_PROPERTIES: set[URIRef] = {
    URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#label"),
    RDFS.label,
    RDFS.comment,
    RDFS.seeAlso,
    DC.title,
    DC.description,
    DC.creator,
    DC.date,
    DCTERMS.created,
    DCTERMS.modified,
    DCTERMS.description,
    SKOS.prefLabel,
    SKOS.altLabel,
    SKOS.note,
    SKOS.definition,
}


def get_short_name(node: Node, graph: Graph) -> str:
    return node.n3(graph.namespace_manager)


def is_annotation_property(prop: Union[URIRef, Node]) -> bool:
    return prop in ANNOTATION_PROPERTIES
